"""Metrics for nonparametric analysis."""

from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

import numpy as np
import scipy.stats

import tea_tasting.config
from tea_tasting.metrics.base import MetricBaseGranular
import tea_tasting.utils


if TYPE_CHECKING:
    from typing import Literal

    import numpy.typing as npt
    import pyarrow as pa


class MannWhitneyUResult(NamedTuple):
    """Result of the analysis using the Mann-Whitney U test.

    Attributes:
        control: ROC AUC for control. Probability that a value from control
            is greater than a value from treatment, plus half the probability
            that they are equal.
        treatment: ROC AUC for treatment. Probability that a value from treatment
            is greater than a value from control, plus half the probability
            that they are equal.
        effect_size: Absolute effect size. Difference between treatment and control
            ROC AUC values.
        pvalue: P-value.
        statistic: Mann-Whitney U statistic.
    """

    control: float
    treatment: float
    effect_size: float
    pvalue: float
    statistic: float


class MannWhitneyU(MetricBaseGranular[MannWhitneyUResult]):  # noqa: D101
    def __init__(
        self,
        column: str,
        *,
        alternative: Literal["two-sided", "greater", "less"] | None = None,
        correction: bool | None = None,
        method: Literal["auto", "asymptotic", "exact"] = "auto",
        nan_policy: Literal["propagate", "omit", "raise"] = "propagate",
    ) -> None:
        """Metric for nonparametric analysis with the Mann-Whitney U test.

        Args:
            column: Metric column name.
            alternative: Alternative hypothesis:

                - `"two-sided"`: the distributions are not equal,
                - `"greater"`: the treatment distribution is stochastically greater
                    than the control distribution,
                - `"less"`: the treatment distribution is stochastically less than
                    the control distribution.

            correction: Whether a continuity correction (1/2) should be applied.
                Only for the asymptotic method.
                Defaults to the global config value (`True`).
            method: Method used for p-value calculation:

                - `"auto"`: exact when sample sizes are small and there are no ties;
                    asymptotic otherwise.
                - `"asymptotic"`: normal approximation with tie correction.
                - `"exact"`: exact p-value calculation.

            nan_policy: Defines how to handle `nan` values:

                - `"propagate"`: return `nan`,
                - `"omit"`: ignore `nan` values,
                - `"raise"`: raise an exception.

        Parameter defaults:
            Defaults for parameters `alternative` and `correction` can be changed using
            the `config_context` and `set_config` functions.
            See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
            reference for details.

        References:
            - [Mann-Whitney U test](https://en.wikipedia.org/wiki/Mann%E2%80%93Whitney_U_test).
            - [scipy.stats.mannwhitneyu &#8212; SciPy Manual](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.mannwhitneyu.html).

        Examples:
            ```pycon
            >>> import tea_tasting as tt

            >>> data = tt.make_users_data(rng=42, n_users=1000)
            >>> experiment = tt.Experiment(
            ...     revenue_auc=tt.MannWhitneyU("revenue"),
            ... )
            >>> result = experiment.analyze(data)
            >>> result
            metric      control treatment rel_effect_size rel_effect_size_ci pvalue
            revenue_auc   0.472     0.528               -             [-, -] 0.0698

            ```

            With specific alternative and method:

            ```pycon
            >>> experiment = tt.Experiment(
            ...     revenue_auc=tt.MannWhitneyU(
            ...         "revenue",
            ...         alternative="greater",
            ...         method="asymptotic",
            ...         correction=False,
            ...     ),
            ... )
            >>> experiment.analyze(data)
            metric      control treatment rel_effect_size rel_effect_size_ci pvalue
            revenue_auc   0.472     0.528               -             [-, -] 0.0349

            ```
        """
        self.column = tea_tasting.utils.check_scalar(column, "column", typ=str)
        self.alternative: Literal["two-sided", "greater", "less"] = (
            tea_tasting.utils.auto_check(alternative, "alternative")
            if alternative is not None
            else tea_tasting.config.get_config("alternative")
        )
        self.correction = (
            tea_tasting.utils.auto_check(correction, "correction")
            if correction is not None
            else tea_tasting.config.get_config("correction")
        )
        self.method = tea_tasting.utils.check_scalar(
            method,
            "method",
            typ=str,
            in_={"auto", "asymptotic", "exact"},
        )
        self.nan_policy: Literal["propagate", "omit", "raise"]
        self.nan_policy = tea_tasting.utils.check_scalar(
            nan_policy,
            "nan_policy",
            typ=str,
            in_={"propagate", "omit", "raise"},
        )

    @property
    def cols(self) -> tuple[str]:
        """Columns to be fetched for a metric analysis."""
        return (self.column,)

    def analyze_granular(
        self,
        control: pa.Table,
        treatment: pa.Table,
    ) -> MannWhitneyUResult:
        """Analyze a metric in an experiment using granular data.

        Args:
            control: Control data.
            treatment: Treatment data.

        Returns:
            Analysis result.
        """
        contr = _select_as_numpy(control, self.column)
        treat = _select_as_numpy(treatment, self.column)
        contr, treat = _handle_nan_policy(contr, treat, self.nan_policy)
        if len(contr) == 0 or len(treat) == 0:
            return MannWhitneyUResult(
                control=float("nan"),
                treatment=float("nan"),
                effect_size=float("nan"),
                pvalue=float("nan"),
                statistic=float("nan"),
            )

        result = scipy.stats.mannwhitneyu(
            treat,
            contr,
            alternative=self.alternative,
            use_continuity=self.correction,
            method=self.method,
        )
        treat_auc = float(result.statistic) / len(contr) / len(treat)
        contr_auc = 1 - treat_auc

        return MannWhitneyUResult(
            control=contr_auc,
            treatment=treat_auc,
            effect_size=treat_auc - contr_auc,
            pvalue=float(result.pvalue),
            statistic=float(result.statistic),
        )


def _select_as_numpy(data: pa.Table, column: str) -> npt.NDArray[np.number]:
    return data[column].combine_chunks().to_numpy(zero_copy_only=False)


def _handle_nan_policy(
    control: npt.NDArray[np.number],
    treatment: npt.NDArray[np.number],
    nan_policy: Literal["propagate", "omit", "raise"],
) -> tuple[npt.NDArray[np.number], npt.NDArray[np.number]]:
    if nan_policy == "omit":
        return control[~np.isnan(control)], treatment[~np.isnan(treatment)]

    if nan_policy == "raise" and (np.isnan(control).any() or np.isnan(treatment).any()):
        raise ValueError("Input contains nan.")

    return control, treatment



from __future__ import annotations

import builtins
import math
import textwrap
from typing import TYPE_CHECKING
import unittest.mock

import numpy as np
import pandas as pd
import pandas.testing
import polars as pl
import polars.testing
import pyarrow as pa
import pytest

import tea_tasting.utils


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


def test_check_scalar_typ() -> None:
    assert tea_tasting.utils.check_scalar(1, typ=int) == 1
    with pytest.raises(TypeError):
        tea_tasting.utils.check_scalar(1, typ=str)

def test_check_scalar_ge() -> None:
    assert tea_tasting.utils.check_scalar(1, ge=1) == 1
    with pytest.raises(ValueError, match="must be >="):
        tea_tasting.utils.check_scalar(1, ge=2)

def test_check_scalar_gt() -> None:
    assert tea_tasting.utils.check_scalar(1, gt=0) == 1
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.check_scalar(1, gt=1)

def test_check_scalar_le() -> None:
    assert tea_tasting.utils.check_scalar(1, le=1) == 1
    with pytest.raises(ValueError, match="must be <="):
        tea_tasting.utils.check_scalar(1, le=0)

def test_check_scalar_lt() -> None:
    assert tea_tasting.utils.check_scalar(1, lt=2) == 1
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.check_scalar(1, lt=1)

def test_check_scalar_ne() -> None:
    assert tea_tasting.utils.check_scalar(1, ne=2) == 1
    with pytest.raises(ValueError, match="must be !="):
        tea_tasting.utils.check_scalar(1, ne=1)

def test_check_scalar_is_in() -> None:
    assert tea_tasting.utils.check_scalar(1, in_={0, 1}) == 1
    with pytest.raises(ValueError, match="must be in"):
        tea_tasting.utils.check_scalar(1, in_={0, 2})


def test_auto_check_alpha() -> None:
    assert tea_tasting.utils.auto_check(0.05, "alpha") == 0.05
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "alpha")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "alpha")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "alpha")

def test_auto_check_alternative() -> None:
    assert tea_tasting.utils.auto_check("two-sided", "alternative") == "two-sided"
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(2, "alternative")
    with pytest.raises(ValueError, match="must be in"):
        tea_tasting.utils.auto_check("2s", "alternative")

def test_auto_check_confidence_level() -> None:
    assert tea_tasting.utils.auto_check(0.95, "confidence_level") == 0.95
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "confidence_level")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "confidence_level")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "confidence_level")

def test_auto_check_correction() -> None:
    assert tea_tasting.utils.auto_check(True, "correction") is True
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "correction")

def test_auto_check_equal_var() -> None:
    assert tea_tasting.utils.auto_check(True, "equal_var") is True
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "equal_var")

def test_auto_check_n_obs() -> None:
    assert tea_tasting.utils.auto_check(2, "n_obs") == 2
    assert tea_tasting.utils.auto_check((2, 3), "n_obs") == (2, 3)
    assert tea_tasting.utils.auto_check(None, "n_obs") is None
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0.5, "n_obs")
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check((0.5, 2), "n_obs")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(1, "n_obs")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check((1, 2), "n_obs")

def test_auto_check_n_resamples() -> None:
    assert tea_tasting.utils.auto_check(1, "n_resamples") == 1
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0, "n_resamples")
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0.5, "n_resamples")

def test_auto_check_power() -> None:
    assert tea_tasting.utils.auto_check(0.8, "power") == 0.8
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "power")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "power")
    with pytest.raises(ValueError, match="must be <"):
        tea_tasting.utils.auto_check(1.0, "power")

def test_auto_check_ratio() -> None:
    assert tea_tasting.utils.auto_check(1.5, "ratio") == 1.5
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check("str", "ratio")
    with pytest.raises(ValueError, match="must be >"):
        tea_tasting.utils.auto_check(0.0, "ratio")

def test_auto_check_rng() -> None:
    generator = np.random.default_rng(42)
    seed_sequence = np.random.SeedSequence(42)
    assert tea_tasting.utils.auto_check(42, "rng") == 42
    assert tea_tasting.utils.auto_check(generator, "rng") is generator
    assert tea_tasting.utils.auto_check(seed_sequence, "rng") is seed_sequence
    assert tea_tasting.utils.auto_check(None, "rng") is None
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0.5, "rng")

def test_auto_check_use_t() -> None:
    assert tea_tasting.utils.auto_check(False, "use_t") is False
    with pytest.raises(TypeError):
        tea_tasting.utils.auto_check(0, "use_t")


def test_format_num() -> None:
    assert tea_tasting.utils.format_num(1.2345) == "1.23"
    assert tea_tasting.utils.format_num(0.9999) == "1.00"
    assert tea_tasting.utils.format_num(1.2345, sig=2) == "1.2"
    assert tea_tasting.utils.format_num(0.12345, pct=True) == "12.3%"
    assert tea_tasting.utils.format_num(None) == "-"
    assert tea_tasting.utils.format_num(float("nan")) == "-"
    assert tea_tasting.utils.format_num(float("inf")) == "∞"
    assert tea_tasting.utils.format_num(float("-inf")) == "-∞"
    assert tea_tasting.utils.format_num(0.00012345) == "1.23e-04"
    assert tea_tasting.utils.format_num(0.00099999) == "1.00e-03"
    assert tea_tasting.utils.format_num(9_999_999) == "9999999"
    assert tea_tasting.utils.format_num(10_000_000) == "1.00e+07"
    assert tea_tasting.utils.format_num(12345, thousands_sep=" ") == "12 345"
    assert tea_tasting.utils.format_num(1.2345, decimal_point=",") == "1,23"
    assert tea_tasting.utils.format_num(0) == "0.00"


def test_get_and_format_num() -> None:
    data: dict[str, object] = {
        "name": "metric",
        "metric": 0.12345,
        "rel_metric": 0.12345,
        "metric_ci_lower": 0.12345,
        "metric_ci_upper": 0.98765,
        "rel_metric_ci_lower": 0.12345,
        "rel_metric_ci_upper": 0.98765,
        "power": 0.87654,
    }
    assert tea_tasting.utils.get_and_format_num(data, "name") == "metric"
    assert tea_tasting.utils.get_and_format_num(data, "metric") == "0.123"
    assert tea_tasting.utils.get_and_format_num(data, "rel_metric") == "12%"
    assert tea_tasting.utils.get_and_format_num(data, "metric_ci") == "[0.123, 0.988]"
    assert tea_tasting.utils.get_and_format_num(data, "rel_metric_ci") == "[12%, 99%]"
    assert tea_tasting.utils.get_and_format_num(data, "power") == "88%"


@pytest.fixture
def dicts_repr() -> tea_tasting.utils.DictsReprMixin:
    class DictsRepr(tea_tasting.utils.DictsReprMixin):
        default_keys = ("name", "a", "b")
        default_text_keys = ("name",)
        def to_dicts(self) -> tuple[dict[str, object], ...]:
            return (
                {"name": "x", "a": 0.12345, "b": 0.23456},
                {"name": "yy", "a": 0.34567, "b": 0.45678},
                {"name": "zzz", "a": 0.56789, "b": 0.67890},
            )
    return DictsRepr()

def test_dicts_repr_mixin_to_arrow(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_arrow().equals(pa.table({
        "name": ("x", "yy", "zzz"),
        "a": (0.12345, 0.34567, 0.56789),
        "b": (0.23456, 0.45678, 0.67890),
    }))

def test_dicts_repr_mixin_to_pandas(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    pandas.testing.assert_frame_equal(
        dicts_repr.to_pandas(),
        pd.DataFrame({
            "name": ("x", "yy", "zzz"),
            "a": (0.12345, 0.34567, 0.56789),
            "b": (0.23456, 0.45678, 0.67890),
        }),
    )

def test_dicts_repr_mixin_to_polars(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    polars.testing.assert_frame_equal(
        dicts_repr.to_polars(),
        pl.DataFrame({
            "name": ("x", "yy", "zzz"),
            "a": (0.12345, 0.34567, 0.56789),
            "b": (0.23456, 0.45678, 0.67890),
        }),
    )

def test_dicts_repr_mixin_to_pretty_dicts_default(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_pretty_dicts() == [
        {"name": "x", "a": "0.123", "b": "0.235"},
        {"name": "yy", "a": "0.346", "b": "0.457"},
        {"name": "zzz", "a": "0.568", "b": "0.679"},
    ]

def test_dicts_repr_mixin_to_pretty_dicts_max_rows(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_pretty_dicts(max_rows=2) == [
        {"name": "x", "a": "0.123", "b": "0.235"},
        {"name": "…", "a": "…", "b": "…"},
        {"name": "zzz", "a": "0.568", "b": "0.679"},
    ]

def test_dicts_repr_mixin_with_defaults(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    new_dicts_repr = dicts_repr.with_defaults(
        keys=("c", "d"),
        max_rows=42,
        align="left",
    )
    assert new_dicts_repr.default_keys == ("c", "d")
    assert new_dicts_repr.default_text_keys == ("name",)
    assert new_dicts_repr.default_max_rows == 42
    assert new_dicts_repr.default_align == "left"

def test_dicts_repr_mixin_with_keys(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.with_keys(("c", "d")).default_keys == ("c", "d")

def test_dicts_repr_mixin_with_max_rows(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.with_max_rows(42).default_max_rows == 42

def test_dicts_repr_mixin_to_string(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_string() == textwrap.dedent("""\
        name     a     b
        x    0.123 0.235
        yy   0.346 0.457
        zzz  0.568 0.679""")

def test_dicts_repr_mixin_to_string_align_left(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_string(align="left") == textwrap.dedent("""\
        name a     b
        x    0.123 0.235
        yy   0.346 0.457
        zzz  0.568 0.679""")

def test_dicts_repr_mixin_to_string_align_right(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_string(align="right") == textwrap.dedent("""\
        name     a     b
           x 0.123 0.235
          yy 0.346 0.457
         zzz 0.568 0.679""")

def test_dicts_repr_mixin_to_html(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_html() == (
        '<table class="dataframe" style="text-align: right;">'
        '<thead><tr><th style="text-align: left;">name</th>'
        '<th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td style="text-align: left;">x</td><td>0.123</td><td>0.235</td></tr>'
        '<tr><td style="text-align: left;">yy</td><td>0.346</td><td>0.457</td></tr>'
        '<tr><td style="text-align: left;">zzz</td><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_to_html_align_left(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_html(align="left") == (
        '<table class="dataframe" style="text-align: left;">'
        '<thead><tr><th>name</th><th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td>x</td><td>0.123</td><td>0.235</td></tr>'
        '<tr><td>yy</td><td>0.346</td><td>0.457</td></tr>'
        '<tr><td>zzz</td><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_to_html_align_right(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_html(align="right") == (
        '<table class="dataframe" style="text-align: right;">'
        '<thead><tr><th>name</th><th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td>x</td><td>0.123</td><td>0.235</td></tr>'
        '<tr><td>yy</td><td>0.346</td><td>0.457</td></tr>'
        '<tr><td>zzz</td><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_to_html_indent(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr.to_html(indent="    ") == textwrap.dedent("""\
        <table class="dataframe" style="text-align: right;">
            <thead>
                <tr>
                    <th style="text-align: left;">name</th>
                    <th>a</th>
                    <th>b</th>
                </tr>
            </thead>
            <tbody>
                <tr>
                    <td style="text-align: left;">x</td>
                    <td>0.123</td>
                    <td>0.235</td>
                </tr>
                <tr>
                    <td style="text-align: left;">yy</td>
                    <td>0.346</td>
                    <td>0.457</td>
                </tr>
                <tr>
                    <td style="text-align: left;">zzz</td>
                    <td>0.568</td>
                    <td>0.679</td>
                </tr>
            </tbody>
        </table>""")

def test_dicts_repr_mixin_mime_marimo(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    mime_type, data = dicts_repr._mime_()
    assert mime_type == "text/html"
    assert data.startswith("<marimo-ui-element")

def test_dicts_repr_mixin_mime_exception(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    def import_side_effect(
        name: str,
        globals_: Mapping[str, object] | None = None,
        locals_: Mapping[str, object] | None = None,
        fromlist: Sequence[str] | None = None,
        level: int = 0,
    ) -> object:
        if name == "marimo":
            raise ImportError("No module named 'marimo'")
        return builtins.__import__(name, globals_, locals_, fromlist, level)
    with unittest.mock.patch("builtins.__import__", side_effect=import_side_effect):
        mime_type, data = dicts_repr._mime_()
    assert mime_type == "text/html"
    assert data.startswith('<table class="dataframe"')

def test_dicts_repr_mixin_repr_html(
    dicts_repr: tea_tasting.utils.DictsReprMixin,
) -> None:
    assert dicts_repr._repr_html_() == (
        '<table class="dataframe" style="text-align: right;">'
        '<thead><tr><th style="text-align: left;">name</th>'
        '<th>a</th><th>b</th></tr></thead>'
        '<tbody>'
        '<tr><td style="text-align: left;">x</td><td>0.123</td><td>0.235</td></tr>'
        '<tr><td style="text-align: left;">yy</td><td>0.346</td><td>0.457</td></tr>'
        '<tr><td style="text-align: left;">zzz</td><td>0.568</td><td>0.679</td></tr>'
        '</tbody></table>'
    )

def test_dicts_repr_mixin_repr(dicts_repr: tea_tasting.utils.DictsReprMixin) -> None:
    assert repr(dicts_repr) == textwrap.dedent("""\
        name     a     b
        x    0.123 0.235
        yy   0.346 0.457
        zzz  0.568 0.679""")

def test_dicts_repr_mixin_str(dicts_repr: tea_tasting.utils.DictsReprMixin) -> None:
    assert str(dicts_repr) == textwrap.dedent("""\
        name     a     b
        x    0.123 0.235
        yy   0.346 0.457
        zzz  0.568 0.679""")


def test_repr_mixin_repr() -> None:
    class Repr(tea_tasting.utils.ReprMixin):
        def __init__(self, a: int, *, b: bool, c: str) -> None:
            self._a = -1
            self.a_ = -1
            self.a = a
            self.b_ = -1
            self.b = b
            self.c = c
    r = Repr(a=1, b=False, c="c")
    assert repr(r) == f"Repr(a=1, b=False, c={'c'!r})"

def test_repr_mixin_repr_obj() -> None:
    class Obj(tea_tasting.utils.ReprMixin):
        ...
    obj = Obj()
    assert repr(obj) == "Obj()"

def test_repr_mixin_repr_pos() -> None:
    class Pos(tea_tasting.utils.ReprMixin):
        def __init__(self, *args: int) -> None:
            self.args = args
    pos = Pos(1, 2, 3)
    with pytest.raises(RuntimeError):
        repr(pos)


def test_div() -> None:
    assert tea_tasting.utils.div(1, 2) == 0.5
    assert tea_tasting.utils.div(1, 0, 3) == 3
    assert tea_tasting.utils.div(1, 0) == float("inf")
    assert math.isnan(tea_tasting.utils.div(0, 0))
    assert math.isnan(tea_tasting.utils.div(-1, 0))


def test_float() -> None:
    typ = tea_tasting.utils.Float
    assert typ(1) + 2 == typ(3)
    assert 1 + typ(2) == typ(3)
    assert typ(1) - 2 == typ(-1)
    assert 1 - typ(2) == typ(-1)
    assert typ(1) / 2 == 0.5
    assert 1 / typ(2) == 0.5
    assert math.isnan(typ(0) / 0)
    assert math.isnan(0 / typ(0))
    assert typ(1) / 0 == float("inf")
    assert 1 / typ(0) == float("inf")
    assert math.isnan(typ(-1) / 0)
    assert math.isnan(-1 / typ(0))
    assert typ(5) // 2 == typ(2)
    assert 5 // typ(2) == typ(2)
    assert typ(5) % 2 == typ(1)
    assert 5 % typ(2) == typ(1)
    assert divmod(typ(5), 2) == (typ(2), typ(1))
    assert divmod(5, typ(2)) == (typ(2), typ(1))
    assert typ(2) ** 3 == typ(8)
    assert 2 ** typ(3) == typ(8)
    assert -typ(1) == typ(-1)
    assert +typ(1) == typ(1)
    assert abs(typ(-1)) == typ(1)
    assert int(typ(1.0)) == 1
    assert float(typ(1.0)) == 1.0
    assert round(typ(11), -1) == typ(10)
    assert math.trunc(typ(1.2)) == typ(1)
    assert math.floor(typ(1.2)) == typ(1)
    assert math.ceil(typ(1.2)) == typ(2)


def test_int() -> None:
    typ = tea_tasting.utils.Int
    assert typ(1) + 2 == typ(3)
    assert 1 + typ(2) == typ(3)
    assert typ(1) - 2 == typ(-1)
    assert 1 - typ(2) == typ(-1)
    assert typ(1) / 2 == 0.5
    assert 1 / typ(2) == 0.5
    assert math.isnan(typ(0) / 0)
    assert math.isnan(0 / typ(0))
    assert typ(1) / 0 == float("inf")
    assert 1 / typ(0) == float("inf")
    assert math.isnan(typ(-1) / 0)
    assert math.isnan(-1 / typ(0))
    assert typ(5) // 2 == typ(2)
    assert 5 // typ(2) == typ(2)
    assert typ(5) % 2 == typ(1)
    assert 5 % typ(2) == typ(1)
    assert divmod(typ(5), 2) == (typ(2), typ(1))
    assert divmod(5, typ(2)) == (typ(2), typ(1))
    assert typ(2) ** 3 == typ(8)
    assert 2 ** typ(3) == typ(8)
    assert -typ(1) == typ(-1)
    assert +typ(1) == typ(1)
    assert abs(typ(-1)) == typ(1)
    assert int(typ(1.0)) == 1
    assert float(typ(1.0)) == 1.0
    assert round(typ(11), -1) == typ(10)
    assert math.trunc(typ(1.2)) == typ(1)
    assert math.floor(typ(1.2)) == typ(1)
    assert math.ceil(typ(1.2)) == typ(1)


def test_numeric() -> None:
    assert isinstance(tea_tasting.utils.numeric(1), tea_tasting.utils.Int)
    assert isinstance(tea_tasting.utils.numeric("1"), tea_tasting.utils.Int)
    assert isinstance(tea_tasting.utils.numeric(1.0), tea_tasting.utils.Float)
    assert isinstance(tea_tasting.utils.numeric("inf"), tea_tasting.utils.Float)



from __future__ import annotations

import math
from typing import NamedTuple
import unittest.mock

import pyarrow as pa
import pyarrow.compute as pc
import pytest

import tea_tasting.aggr
import tea_tasting.config
import tea_tasting.datasets
import tea_tasting.metrics.base
import tea_tasting.metrics.proportion


def append_has_order(data: pa.Table) -> pa.Table:
    return data.append_column(
        "has_order",
        pc.greater(data["orders"], 0).cast(pa.int64()),  # type: ignore
    )

@pytest.fixture
def data_arrow() -> pa.Table:
    return append_has_order(tea_tasting.datasets.make_users_data(n_users=100, rng=42))

@pytest.fixture
def data_aggr(data_arrow: pa.Table) -> dict[object, tea_tasting.aggr.Aggregates]:
    return tea_tasting.aggr.read_aggregates(
        data_arrow,
        group_col="variant",
        has_count=True,
        mean_cols=("has_order",),
        var_cols=(),
        cov_cols=(),
    )


def test_proportion_init_default() -> None:
    metric = tea_tasting.metrics.proportion.Proportion("a")
    assert metric.column == "a"
    assert metric.method == "auto"
    assert metric.alternative == "two-sided"
    assert metric.confidence_level == tea_tasting.config.get_config("confidence_level")
    assert metric.correction == tea_tasting.config.get_config("correction")
    assert metric.equal_var is True

def test_proportion_init_custom() -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "b",
        method="norm",
        alternative="greater",
        confidence_level=0.8,
        correction=False,
        equal_var=True,
    )
    assert metric.column == "b"
    assert metric.method == "norm"
    assert metric.alternative == "greater"
    assert metric.confidence_level == 0.8
    assert metric.correction is False
    assert metric.equal_var is True

def test_proportion_init_config() -> None:
    with tea_tasting.config.config_context(
        alternative="less",
        confidence_level=0.8,
        correction=False,
        equal_var=False,
    ):
        metric = tea_tasting.metrics.proportion.Proportion("a")
    assert metric.column == "a"
    assert metric.method == "auto"
    assert metric.alternative == "less"
    assert metric.confidence_level == 0.8
    assert metric.correction is False
    assert metric.equal_var is True

def test_proportion_init_raises() -> None:
    with pytest.raises(ValueError, match="two-sided"):
        tea_tasting.metrics.proportion.Proportion(
            "a",
            method="log-likelihood",
            alternative="greater",
        )


def test_proportion_aggr_cols() -> None:
    metric = tea_tasting.metrics.proportion.Proportion("a")
    assert metric.aggr_cols == tea_tasting.metrics.base.AggrCols(
        has_count=True,
        mean_cols=("a",),
    )


def test_proportion_analyze_frame(data_arrow: pa.Table) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order")
    result = metric.analyze(data_arrow, 0, 1, variant="variant")
    assert isinstance(result, tea_tasting.metrics.proportion.ProportionResult)

def test_proportion_analyze_auto() -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order")
    with unittest.mock.patch("scipy.stats.barnard_exact") as mock:
        mock.return_value = NamedTuple("Result", (("pvalue", float),))(pvalue=0.1)
        data = append_has_order(tea_tasting.datasets.make_users_data(
            rng=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD - 1,
        ))
        metric.analyze(data, 0, 1, variant="variant")
        mock.assert_called_once()

    with unittest.mock.patch("scipy.stats.barnard_exact") as mock:
        data = append_has_order(tea_tasting.datasets.make_users_data(
            rng=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD,
        ))
        result = metric.analyze(data, 0, 1, variant="variant")
        mock.assert_not_called()
        assert not math.isnan(result.effect_size_ci_lower)
        assert not math.isnan(result.effect_size_ci_upper)
        assert not math.isnan(result.rel_effect_size_ci_lower)
        assert not math.isnan(result.rel_effect_size_ci_upper)

def test_proportion_analyze_barnard(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="barnard")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == pytest.approx(0.3584905660377358)
    assert result.treatment == pytest.approx(0.2765957446808511)
    assert result.effect_size == pytest.approx(-0.08189482135688475)
    assert math.isnan(result.effect_size_ci_lower)
    assert math.isnan(result.effect_size_ci_upper)
    assert result.rel_effect_size == pytest.approx(-0.22844344904815217)
    assert math.isnan(result.rel_effect_size_ci_lower)
    assert math.isnan(result.rel_effect_size_ci_upper)
    assert result.pvalue == pytest.approx(0.530637102766593)

def test_proportion_analyze_barnard_less_pooled(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="barnard", alternative="less", equal_var=True)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.2894571757676508)

def test_proportion_analyze_boschloo(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="boschloo")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.4000169399363184)

def test_proportion_analyze_boschloo_greater(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="boschloo", alternative="greater")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.8249417819623234)

def test_proportion_analyze_fisher(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="fisher")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.3999948918153569)

def test_proportion_analyze_fisher_less(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="fisher", alternative="less")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.2546303652212273)

def test_proportion_analyze_log_likelihood(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="log-likelihood")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.5076294534617167)

def test_proportion_analyze_log_likelihood_no_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="log-likelihood", correction=False)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.37976268998844287)

def test_proportion_analyze_pearson(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="pearson")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.pvalue == pytest.approx(0.5083165530405195)

def test_proportion_analyze_pearson_zero_margin() -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="pearson")
    data = {
        0: tea_tasting.aggr.Aggregates(count_=10, mean_={"has_order": 0}),
        1: tea_tasting.aggr.Aggregates(count_=10, mean_={"has_order": 0}),
    }
    result = metric.analyze(data, 0, 1, variant="variant")  # type: ignore
    assert math.isnan(result.pvalue)

def test_proportion_analyze_norm(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion("has_order", method="norm")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.effect_size_ci_lower == pytest.approx(-0.2851524037206986)
    assert result.effect_size_ci_upper == pytest.approx(0.12136276100692911)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.5884358933441386)
    assert result.rel_effect_size_ci_upper == pytest.approx(0.44643204227351263)
    assert result.pvalue == pytest.approx(0.5083165530405196)

def test_proportion_analyze_norm_unpooled(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", equal_var=False)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.effect_size_ci_lower == pytest.approx(-0.28368859581157896)
    assert result.effect_size_ci_upper == pytest.approx(0.11989895309780946)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.5939973223152473)
    assert result.rel_effect_size_ci_upper == pytest.approx(0.46624528367997886)
    assert result.pvalue == pytest.approx(0.5049066155125612)

def test_proportion_analyze_norm_greater(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", alternative="greater")
    result = metric.analyze(data_aggr, 1, 0, variant="variant")
    assert result.effect_size_ci_lower == pytest.approx(-0.0919114073864812)
    assert result.effect_size_ci_upper == 1
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.25445686244444576)
    assert result.rel_effect_size_ci_upper == float("inf")
    assert result.pvalue == pytest.approx(0.2541582765202598)

def test_proportion_analyze_norm_less(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", alternative="less")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.effect_size_ci_lower == -1
    assert result.effect_size_ci_upper == pytest.approx(0.0919114073864812)
    assert result.rel_effect_size_ci_lower == -1
    assert result.rel_effect_size_ci_upper == pytest.approx(0.3192513062736768)
    assert result.pvalue == pytest.approx(0.2541582765202598)

def test_proportion_analyze_norm_no_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", correction=False)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.effect_size_ci_lower == pytest.approx(-0.26508014358420723)
    assert result.effect_size_ci_upper == pytest.approx(0.10129050087043773)
    assert result.rel_effect_size_ci_lower == pytest.approx(-0.5647346534627081)
    assert result.rel_effect_size_ci_upper == pytest.approx(0.367670355686605)
    assert result.pvalue == pytest.approx(0.3809096823118986)


def test_proportion_analyze_norm_corr_widens_ci(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric_corr = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", correction=True)
    metric_no_corr = tea_tasting.metrics.proportion.Proportion(
        "has_order", method="norm", correction=False)
    result_corr = metric_corr.analyze(data_aggr, 0, 1, variant="variant")
    result_no_corr = metric_no_corr.analyze(data_aggr, 0, 1, variant="variant")
    assert result_corr.effect_size_ci_lower < result_no_corr.effect_size_ci_lower
    assert result_corr.effect_size_ci_upper > result_no_corr.effect_size_ci_upper
    assert (
        result_corr.rel_effect_size_ci_lower <
        result_no_corr.rel_effect_size_ci_lower
    )
    assert (
        result_corr.rel_effect_size_ci_upper >
        result_no_corr.rel_effect_size_ci_upper
    )


def test_proportion_analyze_norm_ci_bounds_limited() -> None:
    metric = tea_tasting.metrics.proportion.Proportion(
        "has_order",
        method="norm",
        correction=False,
    )
    data = {
        0: tea_tasting.aggr.Aggregates(count_=20, mean_={"has_order": 0.05}),
        1: tea_tasting.aggr.Aggregates(count_=20, mean_={"has_order": 1.0}),
    }
    result = metric.analyze(data, 0, 1, variant="variant")  # type: ignore
    assert result.effect_size_ci_lower == pytest.approx(0.640490098153452)
    assert result.effect_size_ci_upper == 1

    result = metric.analyze(data, 1, 0, variant="variant")  # type: ignore
    assert result.effect_size_ci_lower == -1
    assert result.effect_size_ci_upper == pytest.approx(-0.640490098153452)


def test_sample_ratio_init_default() -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio()
    assert metric.ratio == 1
    assert metric.method == "auto"
    assert metric.correction == tea_tasting.config.get_config("correction")

def test_sample_ratio_init_custom() -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio(
        {0: 0.5, 1: 0.5},
        method="norm",
        correction=False,
    )
    assert metric.ratio == {0: 0.5, 1: 0.5}
    assert metric.method == "norm"
    assert metric.correction is False

def test_sample_ratio_init_config() -> None:
    with tea_tasting.config.config_context(correction=False):
        metric = tea_tasting.metrics.proportion.SampleRatio()
    assert metric.correction is False


def test_sample_ratio_aggr_cols() -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio()
    assert metric.aggr_cols == tea_tasting.metrics.base.AggrCols(has_count=True)


def test_sample_ratio_analyze_frame(data_arrow: pa.Table) -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio()
    result = metric.analyze(data_arrow, 0, 1, variant="variant")
    assert isinstance(result, tea_tasting.metrics.proportion.SampleRatioResult)

def test_sample_ratio_analyze_auto() -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio()
    with unittest.mock.patch("scipy.stats.binomtest") as mock:
        mock.return_value = NamedTuple("Result", (("pvalue", float),))(pvalue=0.1)
        data = tea_tasting.datasets.make_users_data(
            rng=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD - 1,
        )
        metric.analyze(data, 0, 1, variant="variant")
        mock.assert_called_once()
    with unittest.mock.patch("scipy.stats.norm.sf") as mock:
        mock.return_value = 0.1
        data = tea_tasting.datasets.make_users_data(
            rng=42,
            n_users=tea_tasting.metrics.proportion._MAX_EXACT_THRESHOLD,
        )
        metric.analyze(data, 0, 1, variant="variant")
        mock.assert_called_once()

def test_sample_ratio_analyze_binom(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio(method="binom")
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.6172994135892521)

def test_sample_ratio_analyze_norm_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio(method="norm", correction=True)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.6170750774519738)

def test_sample_ratio_analyze_norm_no_corr(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio(method="norm", correction=False)
    result = metric.analyze(data_aggr, 0, 1, variant="variant")
    assert result.control == 53
    assert result.treatment == 47
    assert result.pvalue == pytest.approx(0.5485062355001472)

def test_sample_ratio_analyze_aggregates(
    data_aggr: dict[object, tea_tasting.aggr.Aggregates],
) -> None:
    metric = tea_tasting.metrics.proportion.SampleRatio()
    with pytest.raises(NotImplementedError):
        metric.analyze_aggregates(data_aggr[0], data_aggr[1])



"""Metrics for the analysis of proportions."""

from __future__ import annotations

import math
from typing import TYPE_CHECKING, NamedTuple

import numpy as np
import scipy.stats

import tea_tasting.aggr
import tea_tasting.config
import tea_tasting.metrics
from tea_tasting.metrics.base import AggrCols, MetricBaseAggregated
import tea_tasting.utils


if TYPE_CHECKING:
    from typing import Literal

    import ibis.expr.types
    import narwhals.typing


_MAX_EXACT_THRESHOLD = 1000


class ProportionResult(NamedTuple):
    """Result of the analysis of proportions.

    Attributes:
        control: Proportion in control.
        treatment: Proportion in treatment.
        effect_size: Absolute effect size. Difference between the two proportions.
        effect_size_ci_lower: Lower bound of the absolute effect size
            confidence interval.
            Calculated only for the normal approximation (`method="norm"`).
            `nan` for other methods.
        effect_size_ci_upper: Upper bound of the absolute effect size
            confidence interval.
            Calculated only for the normal approximation (`method="norm"`).
            `nan` for other methods.
        rel_effect_size: Relative effect size. Difference between the two proportions,
            divided by the control proportion.
        rel_effect_size_ci_lower: Lower bound of the relative effect size
            confidence interval.
            Calculated only for the normal approximation (`method="norm"`).
            `nan` for other methods.
        rel_effect_size_ci_upper: Upper bound of the relative effect size
            confidence interval.
            Calculated only for the normal approximation (`method="norm"`).
            `nan` for other methods.
        pvalue: P-value.
    """
    control: float
    treatment: float
    effect_size: float
    effect_size_ci_lower: float
    effect_size_ci_upper: float
    rel_effect_size: float
    rel_effect_size_ci_lower: float
    rel_effect_size_ci_upper: float
    pvalue: float


class _ZTestResult(NamedTuple):
    pvalue: float
    effect_size_ci_lower: float
    effect_size_ci_upper: float
    rel_effect_size_ci_lower: float
    rel_effect_size_ci_upper: float


class Proportion(MetricBaseAggregated[ProportionResult]):  # noqa: D101
    def __init__(
        self,
        column: str,
        *,
        method: Literal[
            "auto",
            "barnard",
            "boschloo",
            "fisher",
            "log-likelihood",
            "norm",
            "pearson",
        ] = "auto",
        alternative: Literal["two-sided", "greater", "less"] | None = None,
        confidence_level: float | None = None,
        correction: bool | None = None,
        equal_var: bool = True,
    ) -> None:
        """Metric for the analysis of proportions.

        Args:
            column: Metric column name. Column values should be 0 or 1.
            method: Statistical test used for calculation of p-value:

                - `"auto"`: Barnard's exact test if the total number
                    of observations is < 1000; or normal approximation otherwise.
                - `"barnard"`: Barnard's exact test using the Wald statistic.
                - `"boschloo"`: Boschloo's exact test.
                    Also known as Barnard's exact test with
                    p-value of Fisher's exact test as a statistic.
                - `"fisher"`: Fisher's exact test.
                - `"log-likelihood"`: G-test.
                - `"norm"`: Normal approximation of the binomial distribution.
                    Also known as two-sample proportion Z-test.
                    Confidence intervals for absolute and relative effect size are
                    calculated only for this method.
                - `"pearson"`: Pearson's chi-squared test.

            alternative: Alternative hypothesis:

                - `"two-sided"`: the means are unequal,
                - `"greater"`: the mean in the treatment variant is greater than
                    the mean in the control variant,
                - `"less"`: the mean in the treatment variant is less than the mean
                    in the control variant.

                G-test and Pearson's chi-squared test are always two-sided.
            confidence_level: Confidence level for the confidence intervals.
                Used only with the normal approximation (`method="norm"`).
            correction: If `True`, add continuity correction. Only for
                approximate methods: normal, G-test, and Pearson's chi-squared test.
                Defaults to the global config value (`True`).
            equal_var: Defines whether equal variance is assumed. If `True`,
                pooled variance is used for the calculation of the standard error
                of the difference between two proportions. Only for normal approximation
                and for Barnard's exact test (in the Wald statistic).
                Default is `True`. Global config is ignored as pooled variance is
                optimal for proportion tests.

        Parameter defaults:
            Defaults for parameters `alternative`, `confidence_level`, and `correction`
            can be changed using the `config_context` and `set_config` functions.
            See the [Global configuration](https://tea-tasting.e10v.me/api/config/)
            reference for details.

        References:
            - [Barnard's test](https://en.wikipedia.org/wiki/Barnard%27s_test).
            - [Boschloo's test](https://en.wikipedia.org/wiki/Boschloo%27s_test).
            - [Fisher's exact test](https://en.wikipedia.org/wiki/Fisher%27s_exact_test).
            - [G-test](https://en.wikipedia.org/wiki/G-test).
            - [Two-proportion Z-test](https://en.wikipedia.org/wiki/Two-proportion_Z-test).
            - [Pearson's chi-squared test](https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test).

        Examples:
            ```pycon
            >>> import pyarrow as pa
            >>> import pyarrow.compute as pc
            >>> import tea_tasting as tt

            >>> data = tt.make_users_data(rng=42, n_users=1000)
            >>> data = data.append_column(
            ...     "has_order",
            ...     pc.greater(data["orders"], 0).cast(pa.int64()),
            ... )
            >>> data
            pyarrow.Table
            user: int64
            variant: int64
            sessions: int64
            orders: int64
            revenue: double
            has_order: int64
            ----
            user: [[0,1,2,3,4,...,995,996,997,998,999]]
            variant: [[1,0,1,1,0,...,0,1,0,1,0]]
            sessions: [[1,2,1,2,2,...,1,2,4,2,2]]
            orders: [[0,0,0,2,1,...,1,1,0,0,2]]
            revenue: [[0,0,0,16.57,8.87,...,8.54,11.78,0,0,18.69]]
            has_order: [[0,0,0,1,1,...,1,1,0,0,1]]
            >>> experiment = tt.Experiment(
            ...     prop_users_with_orders=tt.Proportion("has_order"),
            ... )
            >>> experiment.analyze(data)
            metric                 control treatment rel_effect_size rel_effect_size_ci pvalue
            prop_users_with_orders   0.300     0.356             19%       [-1.3%, 43%] 0.0693

            ```

            With specific statistical test:

            ```pycon
            >>> experiment = tt.Experiment(
            ...     prop_users_with_orders=tt.Proportion(
            ...         "has_order",
            ...         method="barnard",
            ...         equal_var=False,
            ...     ),
            ... )
            >>> experiment.analyze(data)
            metric                 control treatment rel_effect_size rel_effect_size_ci pvalue
            prop_users_with_orders   0.300     0.356             19%             [-, -] 0.0620

            ```
        """  # noqa: E501
        self.column = tea_tasting.utils.check_scalar(column, "column", typ=str)
        self.method = tea_tasting.utils.check_scalar(method, "method", typ=str, in_={
            "auto",
            "barnard",
            "boschloo",
            "fisher",
            "log-likelihood",
            "norm",
            "pearson",
        })
        self.alternative: Literal["two-sided", "greater", "less"] = (
            tea_tasting.utils.auto_check(alternative, "alternative")
            if alternative is not None
            else tea_tasting.config.get_config("alternative")
        )
        self.confidence_level = (
            tea_tasting.utils.auto_check(confidence_level, "confidence_level")
            if confidence_level is not None
            else tea_tasting.config.get_config("confidence_level")
        )
        if self.alternative != "two-sided" and method in {"log-likelihood", "pearson"}:
            raise ValueError(
                f"The {method} method supports only two-sided alternative hypothesis.")
        self.correction = (
            tea_tasting.utils.auto_check(correction, "correction")
            if correction is not None
            else tea_tasting.config.get_config("correction")
        )
        self.equal_var = tea_tasting.utils.auto_check(equal_var, "equal_var")


    @property
    def aggr_cols(self) -> AggrCols:
        """Columns to be aggregated for a metric analysis."""
        return tea_tasting.metrics.AggrCols(
            has_count=True,
            mean_cols=(self.column,),
        )


    def analyze_aggregates(
        self,
        control: tea_tasting.aggr.Aggregates,
        treatment: tea_tasting.aggr.Aggregates,
    ) -> ProportionResult:
        """Analyze a metric in an experiment using aggregated statistics.

        Args:
            control: Control data.
            treatment: Treatment data.

        Returns:
            Analysis result.
        """
        control = control.with_zero_div()
        treatment = treatment.with_zero_div()
        p_contr = control.mean(self.column)
        p_treat = treatment.mean(self.column)
        n_contr = control.count()
        n_treat = treatment.count()

        method = self.method
        if method == "auto":
            method = "barnard" if n_contr + n_treat < _MAX_EXACT_THRESHOLD else "norm"

        effect_size_ci_lower = float("nan")
        effect_size_ci_upper = float("nan")
        rel_effect_size_ci_lower = float("nan")
        rel_effect_size_ci_upper = float("nan")

        if method != "norm":
            data = np.empty(shape=(2, 2), dtype=np.int64)
            data[0, 0] = round(n_treat * p_treat)
            data[0, 1] = round(n_contr * p_contr)
            data[1, 0] = n_treat - data[0, 0]
            data[1, 1] = n_contr - data[0, 1]

        if method == "barnard":
            pvalue = scipy.stats.barnard_exact(
                data,  # type: ignore
                alternative=self.alternative,
                pooled=self.equal_var,
            ).pvalue
        elif method == "boschloo":
            pvalue = scipy.stats.boschloo_exact(
                data, alternative=self.alternative).pvalue  # type: ignore
        elif method == "fisher":
            pvalue = scipy.stats.fisher_exact(data, alternative=self.alternative).pvalue  # type: ignore
        elif method in {"log-likelihood", "pearson"}:
            if np.any(data.sum(axis=0) == 0) or np.any(data.sum(axis=1) == 0):  # type: ignore
                pvalue = float("nan")
            else:
                pvalue = scipy.stats.chi2_contingency(
                    data,  # type: ignore
                    correction=self.correction,
                    lambda_=self.method,
                ).pvalue  # type: ignore
        else:  # norm
            norm_result = self._2sample_proportion_ztest(
                p_contr=p_contr,
                p_treat=p_treat,
                n_contr=n_contr,
                n_treat=n_treat,
            )
            pvalue = norm_result.pvalue
            effect_size_ci_lower = norm_result.effect_size_ci_lower
            effect_size_ci_upper = norm_result.effect_size_ci_upper
            rel_effect_size_ci_lower = norm_result.rel_effect_size_ci_lower
            rel_effect_size_ci_upper = norm_result.rel_effect_size_ci_upper

        return ProportionResult(
            control=p_contr,
            treatment=p_treat,
            effect_size=p_treat - p_contr,
            effect_size_ci_lower=effect_size_ci_lower,
            effect_size_ci_upper=effect_size_ci_upper,
            rel_effect_size=p_treat/p_contr - 1,
            rel_effect_size_ci_lower=rel_effect_size_ci_lower,
            rel_effect_size_ci_upper=rel_effect_size_ci_upper,
            pvalue=pvalue,
        )


    def _2sample_proportion_ztest(
        self,
        *,
        p_contr: float,
        p_treat: float,
        n_contr: int,
        n_treat: int,
    ) -> _ZTestResult:
        distr, log_distr = self._2sample_proportion_distr(
            p_contr=p_contr,
            p_treat=p_treat,
            n_contr=n_contr,
            n_treat=n_treat,
        )
        diff = p_treat - p_contr
        p_ratio = p_treat / p_contr
        cc = 0.5 * (1/n_contr + 1/n_treat) if self.correction else 0
        log_cc = cc / p_contr if self.correction else 0

        if self.alternative == "greater":
            pvalue_diff = max(diff - cc, 0) if self.correction and diff > 0 else diff
            pvalue = distr.sf(pvalue_diff)
            q = self.confidence_level

            effect_size_ci_lower = diff + distr.isf(q) - cc
            effect_size_ci_upper = 1

            p_ratio_ci_lower = p_ratio * math.exp(log_distr.isf(q) - log_cc)
            p_ratio_ci_upper = float("inf")
        elif self.alternative == "less":
            pvalue_diff = min(diff + cc, 0) if self.correction and diff < 0 else diff
            pvalue = distr.cdf(pvalue_diff)
            q = self.confidence_level

            effect_size_ci_lower = -1
            effect_size_ci_upper = diff + distr.ppf(q) + cc

            p_ratio_ci_lower = 0
            p_ratio_ci_upper = p_ratio * math.exp(log_distr.ppf(q) + log_cc)
        else:  # two-sided
            pvalue_diff = abs(diff)
            if self.correction and pvalue_diff > 0:
                pvalue_diff = max(pvalue_diff - cc, 0)
            pvalue = 2 * distr.sf(pvalue_diff)
            q = (1 + self.confidence_level) / 2

            effect_half_ci = distr.ppf(q) + cc
            effect_size_ci_lower = diff - effect_half_ci
            effect_size_ci_upper = diff + effect_half_ci

            rel_half_ci = math.exp(log_distr.ppf(q) + log_cc)
            p_ratio_ci_lower = p_ratio / rel_half_ci
            p_ratio_ci_upper = p_ratio * rel_half_ci

        effect_size_ci_lower = max(effect_size_ci_lower, -1)
        effect_size_ci_upper = min(effect_size_ci_upper, 1)

        return _ZTestResult(
            pvalue=pvalue,
            effect_size_ci_lower=effect_size_ci_lower,
            effect_size_ci_upper=effect_size_ci_upper,
            rel_effect_size_ci_lower=p_ratio_ci_lower - 1,
            rel_effect_size_ci_upper=p_ratio_ci_upper - 1,
        )


    def _2sample_proportion_distr(
        self,
        *,
        p_contr: float,
        p_treat: float,
        n_contr: int,
        n_treat: int,
    ) -> tuple[scipy.stats.rv_frozen, scipy.stats.rv_frozen]:
        if self.equal_var:
            p_pooled = (p_contr*n_contr + p_treat*n_treat) / (n_contr + n_treat)
            scale = math.sqrt(p_pooled * (1 - p_pooled) * (1/n_contr + 1/n_treat))
            log_scale = scale / p_pooled
            return scipy.stats.norm(scale=scale), scipy.stats.norm(scale=log_scale)

        scale = math.sqrt(
            p_contr*(1 - p_contr)/n_contr + p_treat*(1 - p_treat)/n_treat)
        log_scale = math.sqrt(
            (1 - p_contr) / (n_contr*p_contr) +
            (1 - p_treat) / (n_treat*p_treat),
        )
        return scipy.stats.norm(scale=scale), scipy.stats.norm(scale=log_scale)


class SampleRatioResult(NamedTuple):
    """Result of the sample ratio mismatch check.

    Attributes:
        control: Number of observations in control.
        treatment: Number of observations in treatment.
        pvalue: P-value.
    """
    control: float
    treatment: float
    pvalue: float


class SampleRatio(MetricBaseAggregated[SampleRatioResult]):  # noqa: D101
    def __init__(
        self,
        ratio: float | int | dict[object, float | int] = 1,
        *,
        method: Literal["auto", "binom", "norm"] = "auto",
        correction: bool | None = None,
    ) -> None:
        """Metric for sample ratio mismatch check.

        Args:
            ratio: Expected ratio of the number of observations in the treatment
                relative to the control.
            method: Statistical test used for calculation of p-value:

                - `"auto"`: Exact binomial test if the total number
                    of observations is < 1000; or normal approximation otherwise.
                - `"binom"`: Exact binomial test.
                - `"norm"`: Normal approximation of the binomial distribution.

            correction: If `True`, add continuity correction.
                Only for normal approximation.
                Defaults to the global config value (`True`).

        Examples:
            ```pycon
            >>> import tea_tasting as tt

            >>> experiment = tt.Experiment(
            ...     sample_ratio=tt.SampleRatio(),
            ... )
            >>> data = tt.make_users_data(rng=42)
            >>> result = experiment.analyze(data)
            >>> result.with_keys(("metric", "control", "treatment", "pvalue"))
            metric       control treatment pvalue
            sample_ratio    2023      1977  0.477

            ```

            Different expected ratio:

            ```pycon
            >>> experiment = tt.Experiment(
            ...     sample_ratio=tt.SampleRatio(0.5),
            ... )
            >>> data = tt.make_users_data(rng=42)
            >>> result = experiment.analyze(data)
            >>> result.with_keys(("metric", "control", "treatment", "pvalue"))
            metric       control treatment    pvalue
            sample_ratio    2023      1977 3.26e-103

            ```
        """
        if isinstance(ratio, dict):
            for val in ratio.values():
                tea_tasting.utils.auto_check(val, "ratio")
        else:
            tea_tasting.utils.auto_check(ratio, "ratio")
        self.ratio = ratio

        self.method = tea_tasting.utils.check_scalar(
            method, "method", typ=str, in_={"auto", "binom", "norm"})
        self.correction = (
            tea_tasting.utils.auto_check(correction, "correction")
            if correction is not None
            else tea_tasting.config.get_config("correction")
        )


    @property
    def aggr_cols(self) -> AggrCols:
        """Columns to be aggregated for a metric analysis."""
        return AggrCols(has_count=True)


    def analyze(
        self,
        data: narwhals.typing.IntoFrame | ibis.expr.types.Table | dict[
            object, tea_tasting.aggr.Aggregates],
        control: object,
        treatment: object,
        variant: str | None = None,
    ) -> SampleRatioResult:
        """Perform a sample ratio mismatch check.

        Args:
            data: Experimental data.
            control: Control variant.
            treatment: Treatment variant.
            variant: Variant column name.

        Returns:
            Analysis result.
        """
        tea_tasting.utils.check_scalar(variant, "variant", typ=str | None)
        aggr = tea_tasting.metrics.aggregate_by_variants(
            data,
            aggr_cols=self.aggr_cols,
            variant=variant,
        )

        k = aggr[treatment].count()
        n = k + aggr[control].count()

        r = (
            self.ratio
            if isinstance(self.ratio, float | int)
            else self.ratio[treatment] / self.ratio[control]
        )
        p = r / (1 + r)

        if (
            self.method == "binom" or
            (self.method == "auto" and n < _MAX_EXACT_THRESHOLD)
        ):
            pvalue = scipy.stats.binomtest(k=int(k), n=int(n), p=p).pvalue
        else:  # norm
            d = abs(k - n*p)
            if self.correction and d > 0:
                d = max(d - 0.5, 0)
            z = d / math.sqrt(n * p * (1 - p))
            pvalue = 2 * scipy.stats.norm.sf(z)

        return SampleRatioResult(
            control=n - k,
            treatment=k,
            pvalue=pvalue,  # type: ignore
        )


    def analyze_aggregates(
        self,
        control: tea_tasting.aggr.Aggregates,
        treatment: tea_tasting.aggr.Aggregates,
    ) -> SampleRatioResult:
        """Stub method for compatibility with the base class."""
        raise NotImplementedError
